@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2022, The PurpleI2P Project
+* Copyright (c) 2013-2023, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -160,6 +160,7 @@ namespace http
 			return true;
 		} else if (url.at(pos_c) == '?') {
 			/* found query part */
+			hasquery = true;
 			path = url.substr(pos_p, pos_c - pos_p);
 			pos_p = pos_c + 1;
 			pos_c = url.find('#', pos_p);
@@ -218,8 +219,10 @@ namespace http
 			}
 		}
 		out += path;
+		if (hasquery) // add query even if it was empty
+			out += "?";
 		if (query != "")
-			out += "?" + query;
+			out += query;
 		if (frag != "")
 			out += "#" + frag;
 		return out;
@@ -347,6 +350,14 @@ namespace http
 		return "";
 	}
 
+	size_t HTTPReq::GetNumHeaders (const std::string& name) const
+	{
+		size_t num = 0;
+		for (auto& it : headers)
+			if (it.first == name) num++;
+		return num;
+	}
+
 	bool HTTPRes::is_chunked() const
 	{
 		auto it = headers.find("Transfer-Encoding");
