@@ -693,10 +693,20 @@ namespace transport
 						SendTerminationAndTerminate (eNTCP2Message3Error);
 						return;
 					}
-					auto addr = ri.GetNTCP2AddressWithStaticKey (m_Establisher->m_RemoteStaticKey);
-					if (!addr)
+					auto addr = m_RemoteEndpoint.address ().is_v4 () ? ri.GetNTCP2V4Address () :
+						(i2p::util::net::IsYggdrasilAddress (m_RemoteEndpoint.address ()) ? ri.GetYggdrasilAddress () : ri.GetNTCP2V6Address ());
+					if (!addr || memcmp (m_Establisher->m_RemoteStaticKey, addr->s, 32))
 					{
-						LogPrint (eLogError, "NTCP2: No NTCP2 address with static key found in SessionConfirmed");
+						LogPrint (eLogError, "NTCP2: Wrong static key in SessionConfirmed");
+						Terminate ();
+						return;
+					}
+					if (addr->IsPublishedNTCP2 () && m_RemoteEndpoint.address () != addr->host &&
+					    (!m_RemoteEndpoint.address ().is_v6 () || (i2p::util::net::IsYggdrasilAddress (m_RemoteEndpoint.address ()) ?
+					     memcmp (m_RemoteEndpoint.address ().to_v6 ().to_bytes ().data () + 1, addr->host.to_v6 ().to_bytes ().data () + 1, 7) : // from the same yggdrasil subnet
+					     memcmp (m_RemoteEndpoint.address ().to_v6 ().to_bytes ().data (), addr->host.to_v6 ().to_bytes ().data (), 8)))) // temporary address
+					{
+						LogPrint (eLogError, "NTCP2: Host mismatch between published address ", addr->host, " and actual endpoint ", m_RemoteEndpoint.address ());
 						Terminate ();
 						return;
 					}
@@ -874,8 +884,20 @@ namespace transport
 			switch (blk)
 			{
 				case eNTCP2BlkDateTime:
+				{
 					LogPrint (eLogDebug, "NTCP2: Datetime");
-				break;
+					if (m_IsEstablished)
+					{
+						uint64_t ts = i2p::util::GetSecondsSinceEpoch ();
+						uint64_t tsA = bufbe32toh (frame + offset);
+						if (tsA < ts - NTCP2_CLOCK_SKEW || tsA > ts + NTCP2_CLOCK_SKEW)
+						{
+							LogPrint (eLogWarning, "NTCP2: Established session time difference ", (int)(ts - tsA), " exceeds clock skew");
+							SendTerminationAndTerminate (eNTCP2ClockSkew);
+						}
+					}
+					break;
+				}
 				case eNTCP2BlkOptions:
 					LogPrint (eLogDebug, "NTCP2: Options");
 				break;
@@ -1120,12 +1142,17 @@ namespace transport
 	{
 		if (!IsEstablished ()) return;
 		auto riLen = i2p::context.GetRouterInfo ().GetBufferLen ();
-		size_t payloadLen = riLen + 4; // 3 bytes block header + 1 byte RI flag
+		size_t payloadLen = riLen + 3 + 1 + 7; // 3 bytes block header + 1 byte RI flag + 7 bytes DateTime
 		m_NextSendBuffer = new uint8_t[payloadLen + 16 + 2 + 64]; // up to 64 bytes padding
-		m_NextSendBuffer[2] = eNTCP2BlkRouterInfo;
-		htobe16buf (m_NextSendBuffer + 3, riLen + 1); // size
-		m_NextSendBuffer[5] = 0; // flag
-		memcpy (m_NextSendBuffer + 6, i2p::context.GetRouterInfo ().GetBuffer (), riLen);
+		// DateTime	block
+		m_NextSendBuffer[2] = eNTCP2BlkDateTime;
+		htobe16buf (m_NextSendBuffer + 3, 4);
+		htobe32buf (m_NextSendBuffer + 5, (i2p::util::GetMillisecondsSinceEpoch () + 500)/1000);
+		// RouterInfo block
+		m_NextSendBuffer[9] = eNTCP2BlkRouterInfo;
+		htobe16buf (m_NextSendBuffer + 10, riLen + 1); // size
+		m_NextSendBuffer[12] = 0; // flag
+		memcpy (m_NextSendBuffer + 13, i2p::context.GetRouterInfo ().GetBuffer (), riLen);
 		// padding block
 		auto paddingSize = CreatePaddingBlock (payloadLen, m_NextSendBuffer + 2 + payloadLen, 64);
 		payloadLen += paddingSize;
