@@ -85,8 +85,7 @@ namespace data
 		if (m_IsRunning)
 		{
 			if (m_PersistProfiles)
-				for (auto& it: m_RouterInfos)
-					it.second->SaveProfile ();
+				SaveProfiles ();
 			DeleteObsoleteProfiles ();
 			m_RouterInfos.clear ();
 			m_Floodfills.clear ();
@@ -153,13 +152,13 @@ namespace data
 				if (!i2p::transport::transports.IsOnline ()) continue; // don't manage netdb when offline
 
 				uint64_t ts = i2p::util::GetSecondsSinceEpoch ();
-				if (ts - lastManageRequest >= 15) // manage requests every 15 seconds
+				if (ts - lastManageRequest >= 15 || ts + 15 < lastManageRequest) // manage requests every 15 seconds
 				{
 					m_Requests.ManageRequests ();
 					lastManageRequest = ts;
 				}
 
-				if (ts - lastSave >= 60) // save routers, manage leasesets and validate subscriptions every minute
+				if (ts - lastSave >= 60 || ts + 60 < lastSave) // save routers, manage leasesets and validate subscriptions every minute
 				{
 					if (lastSave)
 					{
@@ -169,14 +168,17 @@ namespace data
 					lastSave = ts;
 				}
 
-				if (ts - lastDestinationCleanup >= i2p::garlic::INCOMING_TAGS_EXPIRATION_TIMEOUT)
+				if (ts - lastDestinationCleanup >= i2p::garlic::INCOMING_TAGS_EXPIRATION_TIMEOUT ||
+				    ts + i2p::garlic::INCOMING_TAGS_EXPIRATION_TIMEOUT < lastDestinationCleanup)
 				{
 					i2p::context.CleanupDestination ();
 					lastDestinationCleanup = ts;
 				}
 
-				if (ts - lastProfilesCleanup >= (uint64_t)(i2p::data::PEER_PROFILE_AUTOCLEAN_TIMEOUT + profilesCleanupVariance))
+				if (ts - lastProfilesCleanup >= (uint64_t)(i2p::data::PEER_PROFILE_AUTOCLEAN_TIMEOUT + profilesCleanupVariance) ||
+				    ts + i2p::data::PEER_PROFILE_AUTOCLEAN_TIMEOUT < lastProfilesCleanup)
 				{
+					if (m_PersistProfiles) PersistProfiles ();
 					DeleteObsoleteProfiles ();
 					lastProfilesCleanup = ts;
 					profilesCleanupVariance = (rand () % (2 * i2p::data::PEER_PROFILE_AUTOCLEAN_VARIANCE) - i2p::data::PEER_PROFILE_AUTOCLEAN_VARIANCE);
@@ -192,7 +194,8 @@ namespace data
 						if (ts - lastPublish >= NETDB_PUBLISH_CONFIRMATION_TIMEOUT) publish = true;
 					}
 					else if (i2p::context.GetLastUpdateTime () > lastPublish ||
-						ts - lastPublish >= NETDB_PUBLISH_INTERVAL)
+						ts - lastPublish >= NETDB_PUBLISH_INTERVAL ||
+					    ts + NETDB_PUBLISH_INTERVAL < lastPublish)
 					{
 						// new publish
 						m_PublishExcluded.clear ();
@@ -208,7 +211,7 @@ namespace data
 					}
 				}
 
-				if (ts - lastExploratory >= 30) // exploratory every 30 seconds
+				if (ts - lastExploratory >= 30 || ts + 30 < lastExploratory) // exploratory every 30 seconds
 				{
 					auto numRouters = m_RouterInfos.size ();
 					if (!numRouters)
@@ -273,7 +276,24 @@ namespace data
 				bool wasFloodfill = r->IsFloodfill ();
 				{
 					std::unique_lock<std::mutex> l(m_RouterInfosMutex);
-					r->Update (buf, len);
+					if (!r->Update (buf, len))
+					{
+						updated = false;
+						m_Requests.RequestComplete (ident, r);
+						return r;
+					}	
+					if (r->IsUnreachable ())
+					{
+						// delete router as invalid after update
+						m_RouterInfos.erase (ident);
+						if (wasFloodfill)
+						{	
+							std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+							m_Floodfills.remove (r);
+						}	
+						m_Requests.RequestComplete (ident, nullptr);
+						return nullptr;	
+					}	
 				}
 				LogPrint (eLogInfo, "NetDb: RouterInfo updated: ", ident.ToBase64());
 				if (wasFloodfill != r->IsFloodfill ()) // if floodfill status updated
@@ -295,7 +315,7 @@ namespace data
 		else
 		{
 			r = std::make_shared<RouterInfo> (buf, len);
-			if (!r->IsUnreachable () && r->HasValidAddresses () &&
+			if (!r->IsUnreachable () && r->HasValidAddresses () && (!r->IsFloodfill () || !r->GetProfile ()->IsUnreachable ()) &&
 			    i2p::util::GetMillisecondsSinceEpoch () + NETDB_EXPIRATION_TIMEOUT_THRESHOLD*1000LL > r->GetTimestamp ())
 			{
 				bool inserted = false;
@@ -427,7 +447,15 @@ namespace data
 	{
 		auto it = m_RouterInfos.find (ident);
 		if (it != m_RouterInfos.end ())
-			return it->second->SetUnreachable (unreachable);
+		{
+			it->second->SetUnreachable (unreachable);
+			if (unreachable)
+			{
+				auto profile = it->second->GetProfile ();
+				if (profile)
+					profile->Unreachable ();
+			}
+		}
 	}
 
 	void NetDb::Reseed ()
@@ -619,13 +647,19 @@ namespace data
 			std::string ident = it.second->GetIdentHashBase64();
 			if (it.second->IsUpdated ())
 			{
-				it.second->SaveToFile (m_Storage.Path(ident));
+				if (it.second->GetBuffer ())
+				{    
+					// we have something to save
+					it.second->SaveToFile (m_Storage.Path(ident));
+					it.second->SetUnreachable (false);
+					it.second->DeleteBuffer ();
+				}	
 				it.second->SetUpdated (false);
-				it.second->SetUnreachable (false);
-				it.second->DeleteBuffer ();
 				updatedCount++;
 				continue;
 			}
+			if (it.second->GetProfile ()->IsUnreachable ())
+				it.second->SetUnreachable (true);
 			// make router reachable back if too few routers or floodfills
 			if (it.second->IsUnreachable () && (total - deletedCount < NETDB_MIN_ROUTERS || isLowRate ||
 				(it.second->IsFloodfill () && totalFloodfills - deletedFloodfillsCount < NETDB_MIN_FLOODFILLS)))
@@ -671,12 +705,12 @@ namespace data
 				for (auto it = m_RouterInfos.begin (); it != m_RouterInfos.end ();)
 				{
 					if (it->second->IsUnreachable ())
-					{
-						if (m_PersistProfiles) it->second->SaveProfile ();
 						it = m_RouterInfos.erase (it);
-						continue;
-					}
-					++it;
+					else
+					{	
+						it->second->DropProfile ();
+						it++;
+					}	
 				}
 			}
 			// clean up expired floodfills or not floodfills anymore
@@ -686,7 +720,7 @@ namespace data
 					if ((*it)->IsUnreachable () || !(*it)->IsFloodfill ())
 						it = m_Floodfills.erase (it);
 					else
-						++it;
+						it++;
 			}
 		}
 	}
@@ -811,7 +845,7 @@ namespace data
 			{
 				LogPrint (eLogError, "NetDb: Database store message is too long ", len);
 				return;
-			}	
+			}
 			if (!m->from) // unsolicited LS must be received directly
 			{
 				if (storeType == NETDB_STORE_TYPE_LEASESET) // 1
@@ -821,7 +855,7 @@ namespace data
 				}
 				else // all others are considered as LeaseSet2
 				{
-					LogPrint (eLogDebug, "NetDb: Store request: LeaseSet2 of type ", storeType, " for ", ident.ToBase32());
+					LogPrint (eLogDebug, "NetDb: Store request: LeaseSet2 of type ", int(storeType), " for ", ident.ToBase32());
 					updated = AddLeaseSet2 (ident, buf + offset, len - offset, storeType);
 				}
 			}
@@ -1006,11 +1040,10 @@ namespace data
 				lookupType == DATABASE_LOOKUP_TYPE_NORMAL_LOOKUP)
 			{
 				auto router = FindRouter (ident);
-				if (router)
+				if (router && !router->IsUnreachable ())
 				{
 					LogPrint (eLogDebug, "NetDb: Requested RouterInfo ", key, " found");
-					PopulateRouterInfoBuffer (router);
-					if (router->GetBuffer ())
+					if (PopulateRouterInfoBuffer (router))
 						replyMsg = CreateDatabaseStoreMsg (router);
 				}
 			}
@@ -1334,7 +1367,7 @@ namespace data
 		std::unique_lock<std::mutex> l(m_FloodfillsMutex);
 		for (const auto& it: m_Floodfills)
 		{
-			if (!it->IsUnreachable ())
+			if (!it->IsUnreachable () && !it->GetProfile ()->IsUnreachable ())
 			{
 				XORMetric m = destKey ^ it->GetIdentHash ();
 				if (m < minMetric && !excluded.count (it->GetIdentHash ()))
@@ -1365,7 +1398,7 @@ namespace data
 			std::unique_lock<std::mutex> l(m_FloodfillsMutex);
 			for (const auto& it: m_Floodfills)
 			{
-				if (!it->IsUnreachable ())
+				if (!it->IsUnreachable () && !it->GetProfile ()->IsUnreachable ())
 				{
 					XORMetric m = destKey ^ it->GetIdentHash ();
 					if (closeThanUsOnly && ourMetric < m) continue;
@@ -1447,10 +1480,11 @@ namespace data
 		m_LeasesPool.CleanUpMt ();
 	}
 
-	void NetDb::PopulateRouterInfoBuffer (std::shared_ptr<RouterInfo> r)
+	bool NetDb::PopulateRouterInfoBuffer (std::shared_ptr<RouterInfo> r)
 	{
-		if (!r || r->GetBuffer ()) return;
-		r->LoadBuffer (m_Storage.Path (r->GetIdentHashBase64 ()));
+		if (!r) return false;
+		if (r->GetBuffer ()) return true;
+		return r->LoadBuffer (m_Storage.Path (r->GetIdentHashBase64 ()));
 	}
 }
 }
