@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2022, The PurpleI2P Project
+* Copyright (c) 2013-2023, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -238,9 +238,33 @@ namespace proxy {
 		std::string value = params["i2paddresshelper"];
 		len += value.length();
 		b64 = i2p::http::UrlDecode(value);
+
 		// if we need update exists, request formed with update param
-		if (params["update"] == "true") { len += std::strlen("&update=true"); confirm = true; }
-		if (pos != 0 && url.query[pos-1] == '&') { pos--; len++; } // if helper is not only one query option
+		if (params["update"] == "true")
+		{
+			len += std::strlen("&update=true");
+			confirm = true;
+		}
+
+		// if helper is not only one query option and it placed after user's query
+		if (pos != 0 && url.query[pos-1] == '&')
+		{
+			pos--;
+			len++;
+		}
+		// if helper is not only one query option and it placed before user's query
+		else if (pos == 0 && url.query.length () > len && url.query[len] == '&')
+		{
+			// we don't touch the '?' but remove the trailing '&'
+			len++;
+		}
+		else
+		{
+			// there is no more query options, resetting hasquery flag
+			url.hasquery = false;
+		}
+
+		// reset hasquery flag and remove addresshelper from URL
 		url.query.replace(pos, len, "");
 		return true;
 	}
@@ -321,7 +345,7 @@ namespace proxy {
 			if (!m_Addresshelper)
 			{
 				LogPrint(eLogWarning, "HTTPProxy: Addresshelper request rejected");
-				GenericProxyError(tr("Invalid request"), tr("addresshelper is not supported"));
+				GenericProxyError(tr("Invalid request"), tr("Addresshelper is not supported"));
 				return true;
 			}
 
@@ -333,24 +357,51 @@ namespace proxy {
 			}
 			else if (!i2p::client::context.GetAddressBook ().FindAddress (m_RequestURL.host) || m_Confirm)
 			{
+				const std::string referer_raw = m_ClientRequest.GetHeader("Referer");
+				i2p::http::URL referer_url;
+				if (!referer_raw.empty ())
+				{
+					referer_url.parse (referer_raw);
+				}
+				if (m_RequestURL.host != referer_url.host)
+				{
+					if (m_Confirm) // Attempt to forced overwriting by link with "&update=true" from harmful URL
+					{
+						LogPrint (eLogWarning, "HTTPProxy: Address update from addresshelper rejected for ", m_RequestURL.host, " (referer is ", m_RequestURL.host.empty() ? "empty" : "harmful", ")");
+						std::string full_url = m_RequestURL.to_string();
+						std::stringstream ss;
+						ss << tr("Host %s is <font color=red>already in router's addressbook</font>. <b>Be careful: source of this URL may be harmful!</b> Click here to update record: <a href=\"%s%s%s&update=true\">Continue</a>.",
+							m_RequestURL.host.c_str(), full_url.c_str(), (full_url.find('?') != std::string::npos ? "&i2paddresshelper=" : "?i2paddresshelper="), jump.c_str());
+						GenericProxyInfo(tr("Addresshelper forced update rejected"), ss.str());
+					}
+					else // Preventing unauthorized additions to the address book
+					{
+						LogPrint (eLogDebug, "HTTPProxy: Adding address from addresshelper for ", m_RequestURL.host, " (generate refer-base page)");
+						std::string full_url = m_RequestURL.to_string();
+						std::stringstream ss;
+						ss << tr("To add host <b>%s</b> in router's addressbook, click here: <a href=\"%s%s%s\">Continue</a>.",
+							m_RequestURL.host.c_str(), full_url.c_str(), (full_url.find('?') != std::string::npos ? "&i2paddresshelper=" : "?i2paddresshelper="), jump.c_str());
+						GenericProxyInfo(tr("Addresshelper request"), ss.str());
+					}
+					return true; /* request processed */
+				}
+
 				i2p::client::context.GetAddressBook ().InsertAddress (m_RequestURL.host, jump);
 				LogPrint (eLogInfo, "HTTPProxy: Added address from addresshelper for ", m_RequestURL.host);
 				std::string full_url = m_RequestURL.to_string();
 				std::stringstream ss;
-				ss << tr("Host") <<" " << m_RequestURL.host << " " << tr("added to router's addressbook from helper") << ". ";
-				ss << tr("Click here to proceed:") << " <a href=\"" << full_url << "\">" << tr("Continue") << "</a>.";
-				GenericProxyInfo(tr("Addresshelper found"), ss.str());
+				ss << tr("Host %s added to router's addressbook from helper. Click here to proceed: <a href=\"%s\">Continue</a>.",
+					m_RequestURL.host.c_str(), full_url.c_str());
+				GenericProxyInfo(tr("Addresshelper adding"), ss.str());
 				return true; /* request processed */
 			}
 			else
 			{
 				std::string full_url = m_RequestURL.to_string();
 				std::stringstream ss;
-				ss << tr("Host") << " " << m_RequestURL.host << " <font color=red>" << tr("already in router's addressbook") << "</font>. ";
-				ss << tr(/* tr: The "record" means addressbook's record. That message appears when domain was already added to addressbook, but helper link is opened for it. */ "Click here to update record:" );
-				ss << " <a href=\"" << full_url << (full_url.find('?') != std::string::npos ? "&i2paddresshelper=" : "?i2paddresshelper=");
-				ss << jump << "&update=true\">" << tr("Continue") << "</a>.";
-				GenericProxyInfo(tr("Addresshelper found"), ss.str());
+				ss << tr("Host %s is <font color=red>already in router's addressbook</font>. Click here to update record: <a href=\"%s%s%s&update=true\">Continue</a>.",
+					m_RequestURL.host.c_str(), full_url.c_str(), (full_url.find('?') != std::string::npos ? "&i2paddresshelper=" : "?i2paddresshelper="), jump.c_str());
+				GenericProxyInfo(tr("Addresshelper update"), ss.str());
 				return true; /* request processed */
 			}
 		}
@@ -363,7 +414,7 @@ namespace proxy {
 			auto pos = uri.find(":");
 			if(pos == std::string::npos || pos == uri.size() - 1)
 			{
-				GenericProxyError(tr("Invalid request"), tr("invalid request uri"));
+				GenericProxyError(tr("Invalid request"), tr("Invalid request URI"));
 				return true;
 			}
 			else
@@ -423,10 +474,10 @@ namespace proxy {
 				if(m_ProxyURL.parse(m_OutproxyUrl))
 					ForwardToUpstreamProxy();
 				else
-					GenericProxyError(tr("Outproxy failure"), tr("bad outproxy settings"));
+					GenericProxyError(tr("Outproxy failure"), tr("Bad outproxy settings"));
 			} else {
 				LogPrint (eLogWarning, "HTTPProxy: Outproxy failure for ", dest_host, ": no outproxy enabled");
-				std::stringstream ss; ss << tr("Host") << " " << dest_host << " " << tr("not inside I2P network, but outproxy is not enabled");
+				std::stringstream ss; ss << tr("Host %s is not inside I2P network, but outproxy is not enabled", dest_host.c_str());
 				GenericProxyError(tr("Outproxy failure"), ss.str());
 			}
 			return true;
@@ -515,13 +566,13 @@ namespace proxy {
 		else
 		{
 			/* unknown type, complain */
-			GenericProxyError(tr("unknown outproxy url"), m_ProxyURL.to_string());
+			GenericProxyError(tr("Unknown outproxy URL"), m_ProxyURL.to_string());
 		}
 	}
 
 	void HTTPReqHandler::HandleUpstreamProxyResolved(const boost::system::error_code & ec, boost::asio::ip::tcp::resolver::iterator it, ProxyResolvedHandler handler)
 	{
-		if(ec) GenericProxyError(tr("cannot resolve upstream proxy"), ec.message());
+		if(ec) GenericProxyError(tr("Cannot resolve upstream proxy"), ec.message());
 		else handler(*it);
 	}
 
@@ -529,7 +580,7 @@ namespace proxy {
 	{
 		if(!ec) {
 			if(m_RequestURL.host.size() > 255) {
-				GenericProxyError(tr("hostname too long"), m_RequestURL.host);
+				GenericProxyError(tr("Hostname is too long"), m_RequestURL.host);
 				return;
 			}
 			uint16_t port = m_RequestURL.port;
@@ -556,13 +607,13 @@ namespace proxy {
 			reqsize += host.size();
 			m_socks_buf[++reqsize] = 0;
 			boost::asio::async_write(*m_proxysock, boost::asio::buffer(m_socks_buf, reqsize), boost::asio::transfer_all(), std::bind(&HTTPReqHandler::HandleSocksProxySendHandshake, this, std::placeholders::_1, std::placeholders::_2));
-		} else GenericProxyError(tr("cannot connect to upstream socks proxy"), ec.message());
+		} else GenericProxyError(tr("Cannot connect to upstream SOCKS proxy"), ec.message());
 	}
 
 	void HTTPReqHandler::HandleSocksProxySendHandshake(const boost::system::error_code & ec, std::size_t bytes_transferred)
 	{
 		LogPrint(eLogDebug, "HTTPProxy: Upstream SOCKS handshake sent");
-		if(ec) GenericProxyError(tr("Cannot negotiate with socks proxy"), ec.message());
+		if(ec) GenericProxyError(tr("Cannot negotiate with SOCKS proxy"), ec.message());
 		else m_proxysock->async_read_some(boost::asio::buffer(m_socks_buf, 8), std::bind(&HTTPReqHandler::HandleSocksProxyReply, this, std::placeholders::_1, std::placeholders::_2));
 	}
 
@@ -604,7 +655,7 @@ namespace proxy {
 		}
 		else
 		{
-			GenericProxyError(tr("CONNECT error"), tr("Failed to Connect"));
+			GenericProxyError(tr("CONNECT error"), tr("Failed to connect"));
 		}
 	}
 
@@ -615,7 +666,7 @@ namespace proxy {
 			m_send_buf = m_ClientResponse.to_string();
 			boost::asio::async_write(*m_sock, boost::asio::buffer(m_send_buf), boost::asio::transfer_all(), [&] (const boost::system::error_code & ec, std::size_t transferred)
 				{
-					if(ec) GenericProxyError(tr("socks proxy error"), ec.message());
+					if(ec) GenericProxyError(tr("SOCKS proxy error"), ec.message());
 					else HandoverToUpstreamProxy();
 				});
 		} else {
@@ -623,7 +674,7 @@ namespace proxy {
 			LogPrint(eLogDebug, "HTTPProxy: Send ", m_send_buf.size(), " bytes");
 			boost::asio::async_write(*m_proxysock, boost::asio::buffer(m_send_buf), boost::asio::transfer_all(), [&](const boost::system::error_code & ec, std::size_t transferred)
 				{
-					if(ec) GenericProxyError(tr("failed to send request to upstream"), ec.message());
+					if(ec) GenericProxyError(tr("Failed to send request to upstream"), ec.message());
 					else HandoverToUpstreamProxy();
 				});
 		}
@@ -641,18 +692,18 @@ namespace proxy {
 				ss << "error code: ";
 				ss << (int) m_socks_buf[1];
 				std::string msg = ss.str();
-				GenericProxyError(tr("socks proxy error"), msg);
+				GenericProxyError(tr("SOCKS proxy error"), msg);
 			}
 		}
-		else GenericProxyError(tr("No Reply From socks proxy"), ec.message());
+		else GenericProxyError(tr("No reply from SOCKS proxy"), ec.message());
 	}
 
 	void HTTPReqHandler::HandleUpstreamHTTPProxyConnect(const boost::system::error_code & ec)
 	{
 		if(!ec) {
 			LogPrint(eLogDebug, "HTTPProxy: Connected to http upstream");
-			GenericProxyError(tr("cannot connect"), tr("http out proxy not implemented"));
-		} else GenericProxyError(tr("cannot connect to upstream http proxy"), ec.message());
+			GenericProxyError(tr("Cannot connect"), tr("HTTP out proxy not implemented"));
+		} else GenericProxyError(tr("Cannot connect to upstream HTTP proxy"), ec.message());
 	}
 
 	/* will be called after some data received from client */
