@@ -21,7 +21,7 @@ namespace client
 {
 
 	/** set standard socket options */
-	static void I2PTunnelSetSocketOptions(std::shared_ptr<boost::asio::ip::tcp::socket> socket)
+	static void I2PTunnelSetSocketOptions (std::shared_ptr<boost::asio::ip::tcp::socket> socket)
 	{
 		if (socket && socket->is_open())
 		{
@@ -46,10 +46,13 @@ namespace client
 	}
 
 	I2PTunnelConnection::I2PTunnelConnection (I2PService * owner, std::shared_ptr<i2p::stream::Stream> stream,
-		std::shared_ptr<boost::asio::ip::tcp::socket> socket, const boost::asio::ip::tcp::endpoint& target, bool quiet):
-		I2PServiceHandler(owner), m_Socket (socket), m_Stream (stream),
-		m_RemoteEndpoint (target), m_IsQuiet (quiet)
+		const boost::asio::ip::tcp::endpoint& target, bool quiet,
+	    std::shared_ptr<boost::asio::ssl::context> sslCtx):
+		I2PServiceHandler(owner), m_Stream (stream), m_RemoteEndpoint (target), m_IsQuiet (quiet)
 	{
+		m_Socket = std::make_shared<boost::asio::ip::tcp::socket> (owner->GetService ());
+		if (sslCtx)
+			m_SSL = std::make_shared<boost::asio::ssl::stream<boost::asio::ip::tcp::socket&> > (*m_Socket, *sslCtx);
 	}
 
 	I2PTunnelConnection::~I2PTunnelConnection ()
@@ -69,7 +72,7 @@ namespace client
 		Receive ();
 	}
 
-	static boost::asio::ip::address GetLoopbackAddressFor(const i2p::data::IdentHash & addr)
+	boost::asio::ip::address GetLoopbackAddressFor(const i2p::data::IdentHash & addr)
 	{
 		boost::asio::ip::address_v4::bytes_type bytes;
 		const uint8_t * ident = addr;
@@ -80,24 +83,26 @@ namespace client
 	}
 
 #ifdef __linux__
-	static void MapToLoopback(const std::shared_ptr<boost::asio::ip::tcp::socket> & sock, const i2p::data::IdentHash & addr)
+	static void MapToLoopback(std::shared_ptr<boost::asio::ip::tcp::socket> sock, const i2p::data::IdentHash & addr)
 	{
-		// bind to 127.x.x.x address
-		// where x.x.x are first three bytes from ident
-		auto ourIP = GetLoopbackAddressFor(addr);
-		boost::system::error_code ec;
-		sock->bind (boost::asio::ip::tcp::endpoint (ourIP, 0), ec);
-		if (ec)
-			LogPrint (eLogError, "I2PTunnel: Can't bind ourIP to ", ourIP.to_string (), ": ", ec.message ());
-
+		if (sock)
+		{
+			// bind to 127.x.x.x address
+			// where x.x.x are first three bytes from ident
+			auto ourIP = GetLoopbackAddressFor(addr);
+			boost::system::error_code ec;
+			sock->bind (boost::asio::ip::tcp::endpoint (ourIP, 0), ec);
+			if (ec)
+				LogPrint (eLogError, "I2PTunnel: Can't bind ourIP to ", ourIP.to_string (), ": ", ec.message ());
+		}
 	}
 #endif
 
 	void I2PTunnelConnection::Connect (bool isUniqueLocal)
 	{
-		I2PTunnelSetSocketOptions(m_Socket);
 		if (m_Socket)
 		{
+			I2PTunnelSetSocketOptions (m_Socket);
 #ifdef __linux__
 			if (isUniqueLocal && m_RemoteEndpoint.address ().is_v4 () &&
 				m_RemoteEndpoint.address ().to_v4 ().to_bytes ()[0] == 127)
@@ -131,6 +136,7 @@ namespace client
 	void I2PTunnelConnection::Terminate ()
 	{
 		if (Kill()) return;
+		if (m_SSL) m_SSL = nullptr;
 		if (m_Stream)
 		{
 			m_Stream->Close ();
@@ -145,12 +151,17 @@ namespace client
 
 	void I2PTunnelConnection::Receive ()
 	{
-		m_Socket->async_read_some (boost::asio::buffer(m_Buffer, I2P_TUNNEL_CONNECTION_BUFFER_SIZE),
-			std::bind(&I2PTunnelConnection::HandleReceived, shared_from_this (),
-			std::placeholders::_1, std::placeholders::_2));
+		if (m_SSL)
+			m_SSL->async_read_some (boost::asio::buffer(m_Buffer, I2P_TUNNEL_CONNECTION_BUFFER_SIZE),
+				std::bind(&I2PTunnelConnection::HandleReceive, shared_from_this (),
+				std::placeholders::_1, std::placeholders::_2));
+		else
+			m_Socket->async_read_some (boost::asio::buffer(m_Buffer, I2P_TUNNEL_CONNECTION_BUFFER_SIZE),
+				std::bind(&I2PTunnelConnection::HandleReceive, shared_from_this (),
+				std::placeholders::_1, std::placeholders::_2));
 	}
 
-	void I2PTunnelConnection::HandleReceived (const boost::system::error_code& ecode, std::size_t bytes_transferred)
+	void I2PTunnelConnection::HandleReceive (const boost::system::error_code& ecode, std::size_t bytes_transferred)
 	{
 		if (ecode)
 		{
@@ -239,8 +250,12 @@ namespace client
 
 	void I2PTunnelConnection::Write (const uint8_t * buf, size_t len)
 	{
-		boost::asio::async_write (*m_Socket, boost::asio::buffer (buf, len), boost::asio::transfer_all (),
-			std::bind (&I2PTunnelConnection::HandleWrite, shared_from_this (), std::placeholders::_1));
+		if (m_SSL)
+			boost::asio::async_write (*m_SSL, boost::asio::buffer (buf, len), boost::asio::transfer_all (),
+				std::bind (&I2PTunnelConnection::HandleWrite, shared_from_this (), std::placeholders::_1));
+		else
+			boost::asio::async_write (*m_Socket, boost::asio::buffer (buf, len), boost::asio::transfer_all (),
+				std::bind (&I2PTunnelConnection::HandleWrite, shared_from_this (), std::placeholders::_1));
 	}
 
 	void I2PTunnelConnection::HandleConnect (const boost::system::error_code& ecode)
@@ -253,22 +268,45 @@ namespace client
 		else
 		{
 			LogPrint (eLogDebug, "I2PTunnel: Connected");
-			if (m_IsQuiet)
-				StreamReceive ();
+			if (m_SSL)
+				m_SSL->async_handshake (boost::asio::ssl::stream_base::client,
+					std::bind (&I2PTunnelConnection::HandleHandshake, shared_from_this (), std::placeholders::_1));
 			else
-			{
-				// send destination first like received from I2P
-				std::string dest = m_Stream->GetRemoteIdentity ()->ToBase64 ();
-				dest += "\n";
-				if(sizeof(m_StreamBuffer) >= dest.size()) {
-					memcpy (m_StreamBuffer, dest.c_str (), dest.size ());
-				}
-				HandleStreamReceive (boost::system::error_code (), dest.size ());
-			}
-			Receive ();
+				Established ();
 		}
 	}
 
+	void I2PTunnelConnection::HandleHandshake (const boost::system::error_code& ecode)
+	{
+		if (ecode)
+		{
+			LogPrint (eLogError, "I2PTunnel: Handshake error: ", ecode.message ());
+			Terminate ();
+		}
+		else
+		{
+			LogPrint (eLogDebug, "I2PTunnel: SSL connected");
+			Established ();
+		}
+	}
+
+	void I2PTunnelConnection::Established ()
+	{
+		if (m_IsQuiet)
+			StreamReceive ();
+		else
+		{
+			// send destination first like received from I2P
+			std::string dest = m_Stream->GetRemoteIdentity ()->ToBase64 ();
+			dest += "\n";
+			if(sizeof(m_StreamBuffer) >= dest.size()) {
+				memcpy (m_StreamBuffer, dest.c_str (), dest.size ());
+			}
+			HandleStreamReceive (boost::system::error_code (), dest.size ());
+		}
+		Receive ();
+	}
+
 	void I2PClientTunnelConnectionHTTP::Write (const uint8_t * buf, size_t len)
 	{
 		if (m_HeaderSent)
@@ -321,15 +359,24 @@ namespace client
 				m_HeaderSent = true;
 				I2PTunnelConnection::Write ((uint8_t *)m_OutHeader.str ().c_str (), m_OutHeader.str ().length ());
 			}
+			else if (m_OutHeader.tellp () < I2P_TUNNEL_HTTP_MAX_HEADER_SIZE)
+				StreamReceive (); // read more header
+			else
+			{
+				LogPrint (eLogError, "I2PTunnel: HTTP header exceeds max size ", I2P_TUNNEL_HTTP_MAX_HEADER_SIZE);
+				Terminate ();
+			}
 		}
 	}
 
 	I2PServerTunnelConnectionHTTP::I2PServerTunnelConnectionHTTP (I2PService * owner, std::shared_ptr<i2p::stream::Stream> stream,
-		std::shared_ptr<boost::asio::ip::tcp::socket> socket,
-		const boost::asio::ip::tcp::endpoint& target, const std::string& host):
-		I2PTunnelConnection (owner, stream, socket, target), m_Host (host),
+		const boost::asio::ip::tcp::endpoint& target, const std::string& host,
+	    std::shared_ptr<boost::asio::ssl::context> sslCtx):
+		I2PTunnelConnection (owner, stream, target, true, sslCtx), m_Host (host),
 		m_HeaderSent (false), m_ResponseHeaderSent (false), m_From (stream->GetRemoteIdentity ())
 	{
+		if (sslCtx)
+			SSL_set_tlsext_host_name(GetSSL ()->native_handle(), host.c_str ());
 	}
 
 	void I2PServerTunnelConnectionHTTP::Write (const uint8_t * buf, size_t len)
@@ -353,7 +400,7 @@ namespace client
 						// strip up some headers
 						static const std::vector<std::string> excluded // list of excluded headers
 						{
-							"Keep-Alive:", "X-I2P" 
+							"Keep-Alive:", "X-I2P"
 						};
 						bool matched = false;
 						for (const auto& it: excluded)
@@ -375,8 +422,8 @@ namespace client
 							else
 								m_OutHeader << "Connection: close\r\n";
 							connection = true;
-						}	
-						else // forward as is	
+						}
+						else // forward as is
 							m_OutHeader << line << "\n";
 					}
 				}
@@ -404,6 +451,13 @@ namespace client
 				m_HeaderSent = true;
 				I2PTunnelConnection::Write ((uint8_t *)m_OutHeader.str ().c_str (), m_OutHeader.str ().length ());
 			}
+			else if (m_OutHeader.tellp () < I2P_TUNNEL_HTTP_MAX_HEADER_SIZE)
+				StreamReceive (); // read more header
+			else
+			{
+				LogPrint (eLogError, "I2PTunnel: HTTP header exceeds max size ", I2P_TUNNEL_HTTP_MAX_HEADER_SIZE);
+				Terminate ();
+			}
 		}
 	}
 
@@ -460,9 +514,9 @@ namespace client
 	}
 
 	I2PTunnelConnectionIRC::I2PTunnelConnectionIRC (I2PService * owner, std::shared_ptr<i2p::stream::Stream> stream,
-		std::shared_ptr<boost::asio::ip::tcp::socket> socket,
-		const boost::asio::ip::tcp::endpoint& target, const std::string& webircpass):
-		I2PTunnelConnection (owner, stream, socket, target), m_From (stream->GetRemoteIdentity ()),
+		const boost::asio::ip::tcp::endpoint& target, const std::string& webircpass,
+	    std::shared_ptr<boost::asio::ssl::context> sslCtx):
+		I2PTunnelConnection (owner, stream, target, true, sslCtx), m_From (stream->GetRemoteIdentity ()),
 		m_NeedsWebIrc (webircpass.length() ? true : false), m_WebircPass (webircpass)
 	{
 	}
@@ -473,7 +527,8 @@ namespace client
 		if (m_NeedsWebIrc)
 		{
 			m_NeedsWebIrc = false;
-			m_OutPacket << "WEBIRC " << m_WebircPass << " cgiirc " << context.GetAddressBook ().ToAddress (m_From->GetIdentHash ()) << " " << GetSocket ()->local_endpoint ().address () << std::endl;
+			m_OutPacket << "WEBIRC " << m_WebircPass << " cgiirc " << context.GetAddressBook ().ToAddress (m_From->GetIdentHash ())
+				<< " " << GetSocket ()->local_endpoint ().address () << std::endl;
 		}
 
 		m_InPacket.clear ();
@@ -665,7 +720,7 @@ namespace client
 		auto localDestination = GetLocalDestination ();
 		if (localDestination)
 			localDestination->StopAcceptingStreams ();
-		
+
 		ClearHandlers ();
 	}
 
@@ -739,6 +794,17 @@ namespace client
 			LogPrint (eLogError, "I2PTunnel: Can't set local address ", localAddress);
 	}
 
+	void I2PServerTunnel::SetSSL (bool ssl)
+	{
+		if (ssl)
+		{
+			m_SSLCtx = std::make_shared<boost::asio::ssl::context> (boost::asio::ssl::context::sslv23);
+			m_SSLCtx->set_verify_mode(boost::asio::ssl::context::verify_none);
+		}
+		else
+			m_SSLCtx = nullptr;
+	}
+
 	void I2PServerTunnel::Accept ()
 	{
 		if (m_PortDestination)
@@ -779,7 +845,7 @@ namespace client
 
 	std::shared_ptr<I2PTunnelConnection> I2PServerTunnel::CreateI2PConnection (std::shared_ptr<i2p::stream::Stream> stream)
 	{
-		return std::make_shared<I2PTunnelConnection> (this, stream, std::make_shared<boost::asio::ip::tcp::socket> (GetService ()), GetEndpoint ());
+		return std::make_shared<I2PTunnelConnection> (this, stream, GetEndpoint (), true, m_SSLCtx);
 
 	}
 
@@ -793,8 +859,7 @@ namespace client
 
 	std::shared_ptr<I2PTunnelConnection> I2PServerTunnelHTTP::CreateI2PConnection (std::shared_ptr<i2p::stream::Stream> stream)
 	{
-		return std::make_shared<I2PServerTunnelConnectionHTTP> (this, stream,
-			std::make_shared<boost::asio::ip::tcp::socket> (GetService ()), GetEndpoint (), m_Host);
+		return std::make_shared<I2PServerTunnelConnectionHTTP> (this, stream, GetEndpoint (), m_Host, GetSSLCtx ());
 	}
 
 	I2PServerTunnelIRC::I2PServerTunnelIRC (const std::string& name, const std::string& address,
@@ -807,371 +872,7 @@ namespace client
 
 	std::shared_ptr<I2PTunnelConnection> I2PServerTunnelIRC::CreateI2PConnection (std::shared_ptr<i2p::stream::Stream> stream)
 	{
-		return std::make_shared<I2PTunnelConnectionIRC> (this, stream, std::make_shared<boost::asio::ip::tcp::socket> (GetService ()), GetEndpoint (), this->m_WebircPass);
-	}
-
-	void I2PUDPServerTunnel::HandleRecvFromI2P(const i2p::data::IdentityEx& from, uint16_t fromPort, uint16_t toPort, const uint8_t * buf, size_t len)
-	{
-		if (!m_LastSession || m_LastSession->Identity.GetLL()[0] != from.GetIdentHash ().GetLL()[0] || fromPort != m_LastSession->RemotePort)
-		{
-			std::lock_guard<std::mutex> lock(m_SessionsMutex);
-			m_LastSession = ObtainUDPSession(from, toPort, fromPort);
-		}
-		m_LastSession->IPSocket.send_to(boost::asio::buffer(buf, len), m_RemoteEndpoint);
-		m_LastSession->LastActivity = i2p::util::GetMillisecondsSinceEpoch();
-	}
-
-	void I2PUDPServerTunnel::HandleRecvFromI2PRaw (uint16_t, uint16_t, const uint8_t * buf, size_t len)
-	{
-		if (m_LastSession)
-		{
-			m_LastSession->IPSocket.send_to(boost::asio::buffer(buf, len), m_RemoteEndpoint);
-			m_LastSession->LastActivity = i2p::util::GetMillisecondsSinceEpoch();
-		}
-	}
-
-	void I2PUDPServerTunnel::ExpireStale(const uint64_t delta)
-	{
-		std::lock_guard<std::mutex> lock(m_SessionsMutex);
-		uint64_t now = i2p::util::GetMillisecondsSinceEpoch();
-		auto itr = m_Sessions.begin();
-		while(itr != m_Sessions.end()) {
-			if(now - (*itr)->LastActivity >= delta )
-				itr = m_Sessions.erase(itr);
-			else
-				++itr;
-		}
-	}
-
-	void I2PUDPClientTunnel::ExpireStale(const uint64_t delta)
-	{
-		std::lock_guard<std::mutex> lock(m_SessionsMutex);
-		uint64_t now = i2p::util::GetMillisecondsSinceEpoch();
-		std::vector<uint16_t> removePorts;
-		for (const auto & s : m_Sessions) {
-			if (now - s.second->second >= delta)
-				removePorts.push_back(s.first);
-		}
-		for(auto port : removePorts) {
-			m_Sessions.erase(port);
-		}
-	}
-
-	UDPSessionPtr I2PUDPServerTunnel::ObtainUDPSession(const i2p::data::IdentityEx& from, uint16_t localPort, uint16_t remotePort)
-	{
-		auto ih = from.GetIdentHash();
-		for (auto & s : m_Sessions )
-		{
-			if (s->Identity.GetLL()[0] == ih.GetLL()[0] && remotePort == s->RemotePort)
-			{
-				/** found existing session */
-				LogPrint(eLogDebug, "UDPServer: Found session ", s->IPSocket.local_endpoint(), " ", ih.ToBase32());
-				return s;
-			}
-		}
-		boost::asio::ip::address addr;
-		/** create new udp session */
-		if(m_IsUniqueLocal && m_LocalAddress.is_loopback())
-		{
-			auto ident = from.GetIdentHash();
-			addr = GetLoopbackAddressFor(ident);
-		}
-		else
-			addr = m_LocalAddress;
-		boost::asio::ip::udp::endpoint ep(addr, 0);
-		m_Sessions.push_back(std::make_shared<UDPSession>(ep, m_LocalDest, m_RemoteEndpoint, &ih, localPort, remotePort));
-		auto & back = m_Sessions.back();
-		return back;
-	}
-
-	UDPSession::UDPSession(boost::asio::ip::udp::endpoint localEndpoint,
-		const std::shared_ptr<i2p::client::ClientDestination> & localDestination,
-		boost::asio::ip::udp::endpoint endpoint, const i2p::data::IdentHash * to,
-		uint16_t ourPort, uint16_t theirPort) :
-		m_Destination(localDestination->GetDatagramDestination()),
-		IPSocket(localDestination->GetService(), localEndpoint),
-		SendEndpoint(endpoint),
-		LastActivity(i2p::util::GetMillisecondsSinceEpoch()),
-		LocalPort(ourPort),
-		RemotePort(theirPort)
-	{
-		IPSocket.set_option (boost::asio::socket_base::receive_buffer_size (I2P_UDP_MAX_MTU ));
-		memcpy(Identity, to->data(), 32);
-		Receive();
-	}
-
-	void UDPSession::Receive()
-	{
-		LogPrint(eLogDebug, "UDPSession: Receive");
-		IPSocket.async_receive_from(boost::asio::buffer(m_Buffer, I2P_UDP_MAX_MTU),
-			FromEndpoint, std::bind(&UDPSession::HandleReceived, this, std::placeholders::_1, std::placeholders::_2));
-	}
-
-	void UDPSession::HandleReceived(const boost::system::error_code & ecode, std::size_t len)
-	{
-		if(!ecode)
-		{
-			LogPrint(eLogDebug, "UDPSession: Forward ", len, "B from ", FromEndpoint);
-			auto ts = i2p::util::GetMillisecondsSinceEpoch();
-			auto session = m_Destination->GetSession (Identity);
-			if (ts > LastActivity + I2P_UDP_REPLIABLE_DATAGRAM_INTERVAL)
-				m_Destination->SendDatagram(session, m_Buffer, len, LocalPort, RemotePort);
-			else
-				m_Destination->SendRawDatagram(session, m_Buffer, len, LocalPort, RemotePort);
-			size_t numPackets = 0;
-			while (numPackets < i2p::datagram::DATAGRAM_SEND_QUEUE_MAX_SIZE)
-			{
-				boost::system::error_code ec;
-				size_t moreBytes = IPSocket.available(ec);
-				if (ec || !moreBytes) break;
-				len = IPSocket.receive_from (boost::asio::buffer (m_Buffer, I2P_UDP_MAX_MTU), FromEndpoint, 0, ec);
-				m_Destination->SendRawDatagram (session, m_Buffer, len, LocalPort, RemotePort);
-				numPackets++;
-			}
-			if (numPackets > 0)
-				LogPrint(eLogDebug, "UDPSession: Forward more ", numPackets, "packets B from ", FromEndpoint);
-			m_Destination->FlushSendQueue (session);
-			LastActivity = ts;
-			Receive();
-		}
-		else
-			LogPrint(eLogError, "UDPSession: ", ecode.message());
-	}
-
-	I2PUDPServerTunnel::I2PUDPServerTunnel (const std::string & name, std::shared_ptr<i2p::client::ClientDestination> localDestination,
-		boost::asio::ip::address localAddress, boost::asio::ip::udp::endpoint forwardTo, uint16_t port, bool gzip) :
-		m_IsUniqueLocal (true), m_Name (name), m_LocalAddress (localAddress), 
-		m_RemoteEndpoint (forwardTo), m_LocalDest (localDestination), m_Gzip (gzip)
-	{
-	}
-
-	I2PUDPServerTunnel::~I2PUDPServerTunnel ()
-	{
-		Stop ();
-	}
-
-	void I2PUDPServerTunnel::Start ()
-	{
-		m_LocalDest->Start ();
-
-		auto dgram = m_LocalDest->CreateDatagramDestination (m_Gzip);
-		dgram->SetReceiver (std::bind (&I2PUDPServerTunnel::HandleRecvFromI2P, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4, std::placeholders::_5));
-		dgram->SetRawReceiver (std::bind (&I2PUDPServerTunnel::HandleRecvFromI2PRaw, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4));
-	}
-
-	void I2PUDPServerTunnel::Stop ()
-	{
-		auto dgram = m_LocalDest->GetDatagramDestination ();
-		if (dgram) dgram->ResetReceiver ();
-	}
-
-	std::vector<std::shared_ptr<DatagramSessionInfo> > I2PUDPServerTunnel::GetSessions ()
-	{
-		std::vector<std::shared_ptr<DatagramSessionInfo> > sessions;
-		std::lock_guard<std::mutex> lock (m_SessionsMutex);
-
-		for (UDPSessionPtr s: m_Sessions)
-		{
-			if (!s->m_Destination) continue;
-			auto info = s->m_Destination->GetInfoForRemote (s->Identity);
-			if (!info) continue;
-
-			auto sinfo = std::make_shared<DatagramSessionInfo> ();
-			sinfo->Name = m_Name;
-			sinfo->LocalIdent = std::make_shared<i2p::data::IdentHash> (m_LocalDest->GetIdentHash ().data ());
-			sinfo->RemoteIdent = std::make_shared<i2p::data::IdentHash> (s->Identity.data ());
-			sinfo->CurrentIBGW = info->IBGW;
-			sinfo->CurrentOBEP = info->OBEP;
-			sessions.push_back (sinfo);
-		}
-		return sessions;
-	}
-
-	I2PUDPClientTunnel::I2PUDPClientTunnel (const std::string & name, const std::string &remoteDest,
-		boost::asio::ip::udp::endpoint localEndpoint,
-		std::shared_ptr<i2p::client::ClientDestination> localDestination,
-		uint16_t remotePort, bool gzip) :
-		m_Name (name), m_RemoteDest (remoteDest), m_LocalDest (localDestination), m_LocalEndpoint (localEndpoint),
-		m_RemoteIdent (nullptr), m_ResolveThread (nullptr), m_LocalSocket (nullptr), RemotePort (remotePort),
-		m_LastPort (0), m_cancel_resolve (false), m_Gzip (gzip)
-	{
-	}
-
-	I2PUDPClientTunnel::~I2PUDPClientTunnel ()
-	{
-		Stop ();
-	}
-
-	void I2PUDPClientTunnel::Start ()
-	{
-		// Reset flag in case of tunnel reload
-		if (m_cancel_resolve) m_cancel_resolve = false;
-
-		m_LocalSocket.reset (new boost::asio::ip::udp::socket (m_LocalDest->GetService (), m_LocalEndpoint));
-		m_LocalSocket->set_option (boost::asio::socket_base::receive_buffer_size (I2P_UDP_MAX_MTU));
-		m_LocalSocket->set_option (boost::asio::socket_base::reuse_address (true));
-
-		auto dgram = m_LocalDest->CreateDatagramDestination (m_Gzip);
-		dgram->SetReceiver (std::bind (&I2PUDPClientTunnel::HandleRecvFromI2P, this,
-			std::placeholders::_1, std::placeholders::_2,
-			std::placeholders::_3, std::placeholders::_4,
-			std::placeholders::_5));
-		dgram->SetRawReceiver (std::bind (&I2PUDPClientTunnel::HandleRecvFromI2PRaw, this,
-			std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4));
-
-		m_LocalDest->Start ();
-		if (m_ResolveThread == nullptr)
-			m_ResolveThread = new std::thread (std::bind (&I2PUDPClientTunnel::TryResolving, this));
-		RecvFromLocal ();
-	}
-
-	void I2PUDPClientTunnel::Stop ()
-	{
-		auto dgram = m_LocalDest->GetDatagramDestination ();
-		if (dgram) dgram->ResetReceiver ();
-		m_cancel_resolve = true;
-
-		m_Sessions.clear();
-
-		if(m_LocalSocket && m_LocalSocket->is_open ())
-			m_LocalSocket->close ();
-
-		if(m_ResolveThread)
-		{
-			m_ResolveThread->join ();
-			delete m_ResolveThread;
-			m_ResolveThread = nullptr;
-		}
-		if (m_RemoteIdent)
-		{
-			delete m_RemoteIdent;
-			m_RemoteIdent = nullptr;
-		}
-	}
-
-	void I2PUDPClientTunnel::RecvFromLocal ()
-	{
-		m_LocalSocket->async_receive_from (boost::asio::buffer (m_RecvBuff, I2P_UDP_MAX_MTU),
-			m_RecvEndpoint, std::bind (&I2PUDPClientTunnel::HandleRecvFromLocal, this, std::placeholders::_1, std::placeholders::_2));
-	}
-
-	void I2PUDPClientTunnel::HandleRecvFromLocal (const boost::system::error_code & ec, std::size_t transferred)
-	{
-		if (m_cancel_resolve) {
-			LogPrint (eLogDebug, "UDP Client: Ignoring incomming data: stopping");
-			return;
-		}
-		if (ec) {
-			LogPrint (eLogError, "UDP Client: Reading from socket error: ", ec.message (), ". Restarting listener...");
-			RecvFromLocal (); // Restart listener and continue work
-			return;
-		}
-		if (!m_RemoteIdent) {
-			LogPrint (eLogWarning, "UDP Client: Remote endpoint not resolved yet");
-			RecvFromLocal ();
-			return; // drop, remote not resolved
-		}
-		auto remotePort = m_RecvEndpoint.port ();
-		if (!m_LastPort || m_LastPort != remotePort)
-		{
-			auto itr = m_Sessions.find (remotePort);
-			if (itr != m_Sessions.end ())
-				m_LastSession = itr->second;
-			else
-			{
-				m_LastSession = std::make_shared<UDPConvo> (boost::asio::ip::udp::endpoint (m_RecvEndpoint), 0);
-				m_Sessions.emplace (remotePort, m_LastSession);
-			}
-			m_LastPort = remotePort;
-		}
-		// send off to remote i2p destination
-		auto ts = i2p::util::GetMillisecondsSinceEpoch ();
-		LogPrint (eLogDebug, "UDP Client: Send ", transferred, " to ", m_RemoteIdent->ToBase32 (), ":", RemotePort);
-		auto session = m_LocalDest->GetDatagramDestination ()->GetSession (*m_RemoteIdent);
-		if (ts > m_LastSession->second + I2P_UDP_REPLIABLE_DATAGRAM_INTERVAL)
-			m_LocalDest->GetDatagramDestination ()->SendDatagram (session, m_RecvBuff, transferred, remotePort, RemotePort);
-		else
-			m_LocalDest->GetDatagramDestination ()->SendRawDatagram (session, m_RecvBuff, transferred, remotePort, RemotePort);
-		size_t numPackets = 0;
-		while (numPackets < i2p::datagram::DATAGRAM_SEND_QUEUE_MAX_SIZE)
-		{
-			boost::system::error_code ec;
-			size_t moreBytes = m_LocalSocket->available (ec);
-			if (ec || !moreBytes) break;
-			transferred = m_LocalSocket->receive_from (boost::asio::buffer (m_RecvBuff, I2P_UDP_MAX_MTU), m_RecvEndpoint, 0, ec);
-			remotePort = m_RecvEndpoint.port ();
-			// TODO: check remotePort
-			m_LocalDest->GetDatagramDestination ()->SendRawDatagram (session, m_RecvBuff, transferred, remotePort, RemotePort);
-			numPackets++;
-		}
-		if (numPackets)
-			LogPrint (eLogDebug, "UDP Client: Sent ", numPackets, " more packets to ", m_RemoteIdent->ToBase32 ());
-		m_LocalDest->GetDatagramDestination ()->FlushSendQueue (session);
-
-		// mark convo as active
-		if (m_LastSession)
-			m_LastSession->second = ts;
-		RecvFromLocal ();
-	}
-
-	std::vector<std::shared_ptr<DatagramSessionInfo> > I2PUDPClientTunnel::GetSessions ()
-	{
-		// TODO: implement
-		std::vector<std::shared_ptr<DatagramSessionInfo> > infos;
-		return infos;
-	}
-
-	void I2PUDPClientTunnel::TryResolving ()
-	{
-		i2p::util::SetThreadName ("UDP Resolver");
-		LogPrint (eLogInfo, "UDP Tunnel: Trying to resolve ", m_RemoteDest);
-
-		std::shared_ptr<const Address> addr;
-		while (!(addr = context.GetAddressBook().GetAddress(m_RemoteDest)) && !m_cancel_resolve)
-		{
-			LogPrint (eLogWarning, "UDP Tunnel: Failed to lookup ", m_RemoteDest);
-			std::this_thread::sleep_for (std::chrono::seconds (1));
-		}
-		if (m_cancel_resolve)
-		{
-			LogPrint(eLogError, "UDP Tunnel: Lookup of ", m_RemoteDest, " was cancelled");
-			return;
-		}
-		if (!addr || !addr->IsIdentHash ())
-		{
-			LogPrint (eLogError, "UDP Tunnel: ", m_RemoteDest, " not found");
-			return;
-		}
-		m_RemoteIdent = new i2p::data::IdentHash;
-		*m_RemoteIdent = addr->identHash;
-		LogPrint(eLogInfo, "UDP Tunnel: Resolved ", m_RemoteDest, " to ", m_RemoteIdent->ToBase32 ());
-	}
-
-	void I2PUDPClientTunnel::HandleRecvFromI2P (const i2p::data::IdentityEx& from, uint16_t fromPort, uint16_t toPort, const uint8_t * buf, size_t len)
-	{
-		if (m_RemoteIdent && from.GetIdentHash() == *m_RemoteIdent)
-			HandleRecvFromI2PRaw (fromPort, toPort, buf, len);
-		else
-			LogPrint(eLogWarning, "UDP Client: Unwarranted traffic from ", from.GetIdentHash().ToBase32 ());
-	}
-
-	void I2PUDPClientTunnel::HandleRecvFromI2PRaw (uint16_t fromPort, uint16_t toPort, const uint8_t * buf, size_t len)
-	{
-		auto itr = m_Sessions.find (toPort);
-		// found convo ?
-		if (itr != m_Sessions.end ())
-		{
-			// found convo
-			if (len > 0)
-			{
-				LogPrint (eLogDebug, "UDP Client: Got ", len, "B from ", m_RemoteIdent ? m_RemoteIdent->ToBase32 () : "");
-				m_LocalSocket->send_to (boost::asio::buffer (buf, len), itr->second->first);
-				// mark convo as active
-				itr->second->second = i2p::util::GetMillisecondsSinceEpoch ();
-			}
-		}
-		else
-			LogPrint (eLogWarning, "UDP Client: Not tracking udp session using port ", (int) toPort);
+		return std::make_shared<I2PTunnelConnectionIRC> (this, stream, GetEndpoint (), m_WebircPass, GetSSLCtx ());
 	}
 }
 }
